// g++ -Wall -Wpedantic -Wextra -std=c++23 -march=haswell -O3 transform.cpp && ./a.out

#include <iostream>
#include <vector>
#include <algorithm>
#include <time.h>
#include <random>

void print_vector(std::vector<int> v) {
    for (std::vector<int>::iterator i = v.begin(); i != v.end(); i++) {
        std::cout << ' ' << *i;
    }
    std::cout << '\n';
}

int add_one(int n) {
    return n + 1;
}

void transform_example() {
    std::vector<int> v{ 10, 20, 30, 40 };
    std::transform(v.cbegin(), v.cend(), v.begin(), add_one);
    std::transform(v.cbegin(), v.cend(), v.begin(), [](int n) {
        return n + 1;
    });
    print_vector(v);
}

#define LENGTH 10000000
void transform_timing() {
    struct timespec st, en;
    // What a delight the STL is... https://stackoverflow.com/a/23143753
    std::random_device rnd_device;
    std::mt19937 mersenne_engine {rnd_device()};
    std::uniform_int_distribution<int> dist {1, 10};
    auto rand = [&dist, &mersenne_engine]() {
        return dist(mersenne_engine);
    };

    std::vector<int> v(LENGTH);
    std::generate(v.begin(), v.end(), rand);

    clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &st);
    std::transform(v.cbegin(), v.cend(), v.begin(), [](int n) {
        return n + 1;
    });
    /*
    for (int i = 0; i < LENGTH; i++) {
        v[i] += 1;
    }
    */
    clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &en);

    double posix_dur = 1000.0 * en.tv_sec + 1e-6 * en.tv_nsec - (1000.0 * st.tv_sec + 1e-6 * st.tv_nsec);

    std::cout << std::accumulate(v.begin(), v.end(), 0) << '\n';
    std::cout << posix_dur << '\n';
}

int main() {
    transform_example();
    transform_timing();

    return 0;
}